/*
 * libwebmdec: a decoder library for WebM audio/video streams
 * Copyright (c) 2014 Andrew Church <achurch@achurch.org>
 *
 * This software may be copied and redistributed under certain conditions;
 * see the file "COPYING" in the source code distribution for details.
 * NO WARRANTY is provided with this software.
 */

#ifdef DECODE_AUDIO  /* To the end of the file. */

#include "include/webmdec.h"
#include "include/internal.h"

#include <stdlib.h>
#include <string.h>

#include <nogg.h>

/*************************************************************************/
/************************ Decoder data structure *************************/
/*************************************************************************/

struct vorbis_decoder_t {
    /* libnogg handle. */
    vorbis_t *nogg;

    /* Buffer for generating Ogg page headers. */
    uint8_t page_header[27+255];
    /* Current Ogg page number. */
    uint32_t page_number;
    /* Pointer to packet data. */
    const uint8_t *packet;
    /* Length of packet data. */
    int32_t packet_len;
    /* Offset of next byte to send to the decoder.  Negative values
     * indicate header bytes, counting from the end of page_header[]. */
    int32_t next_byte;

    /* Pointer to and length of next packet's data.  This is used during
     * initialization, when we need to have both the ident and setup
     * packets ready to hand off to libnogg. */
    const uint8_t *next_packet;
    int32_t next_packet_len;

    /* PCM data not yet returned to the caller. */
    float *pcm;
    int32_t num_samples;
};

/*************************************************************************/
/**************************** Helper routines ****************************/
/*************************************************************************/

/**
 * gen_page_header:  Generate an Ogg page header for the packet whose
 * length is stored in decoder->packet_len.
 *
 * [Parameters]
 *     decoder: Decoder handle.
 *     bos: True to set the beginning-of-stream flag in the header.
 */
static void gen_page_header(vorbis_decoder_t *decoder, int bos)
{
    const int num_segments = (decoder->packet_len / 255) + 1;
    const int header_len = 27 + num_segments;

    uint8_t *header =
        &decoder->page_header[sizeof(decoder->page_header) - header_len];
    header[ 0] = 'O';
    header[ 1] = 'g';
    header[ 2] = 'g';
    header[ 3] = 'S';
    header[ 4] = 0;
    header[ 5] = bos ? 2 : 0;
    header[ 6] = 0xFF;  // Don't bother counting samples.
    header[ 7] = 0xFF;
    header[ 8] = 0xFF;
    header[ 9] = 0xFF;
    header[10] = 0xFF;
    header[11] = 0xFF;
    header[12] = 0xFF;
    header[13] = 0xFF;
    header[14] = 0x01;
    header[15] = 0x00;
    header[16] = 0x00;
    header[17] = 0x00;
    header[18] = decoder->page_number >>  0;
    header[19] = decoder->page_number >>  8;
    header[20] = decoder->page_number >> 16;
    header[21] = decoder->page_number >> 24;
    header[22] = 0x00;
    header[23] = 0x00;
    header[24] = 0x00;
    header[25] = 0x00;
    header[26] = num_segments;
    memset(&header[27], 255, num_segments - 1);
    header[27 + (num_segments - 1)] = decoder->packet_len % 255;

    decoder->next_byte = -(header_len);
    decoder->page_number++;
}

/*-----------------------------------------------------------------------*/

/**
 * nogg_read:  Read callback for libnogg.
 */
static int32_t nogg_read(void *opaque, void *buffer, int32_t length)
{
    vorbis_decoder_t *decoder = opaque;

    const int32_t orig_length = length;
    if (length > decoder->packet_len - decoder->next_byte) {
        length = decoder->packet_len - decoder->next_byte;
    }

    int32_t copied = 0;
    if (decoder->next_byte < 0) {
        int32_t to_copy = length;
        if (to_copy > -(decoder->next_byte)) {
            to_copy = -(decoder->next_byte);
        }
        memcpy(buffer, (decoder->page_header + sizeof(decoder->page_header)
                        + decoder->next_byte), to_copy);
        buffer = (char *)buffer + to_copy;
        decoder->next_byte += to_copy;
        copied = to_copy;
    }
    memcpy(buffer, decoder->packet + decoder->next_byte, length - copied);
    buffer = (char *)buffer + (length - copied);
    decoder->next_byte += (length - copied);

    if (decoder->next_packet && decoder->next_byte >= decoder->packet_len) {
        decoder->packet = decoder->next_packet;
        decoder->packet_len = decoder->next_packet_len;
        decoder->next_packet = NULL;
        decoder->next_packet_len = 0;
        gen_page_header(decoder, 0);
        length += nogg_read(decoder, buffer, orig_length - length);
    }

    return length;
}

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

vorbis_decoder_t *vorbis_decoder_create(
    const unsigned char *id_header, int id_length,
    const unsigned char *setup_header, int setup_length)
{
    if (id_header == NULL
     || id_length < 7
     || memcmp(id_header, "\1vorbis", 7) != 0
     || setup_header == NULL
     || setup_length < 7
     || memcmp(setup_header, "\5vorbis", 7) != 0)
    {
        goto error_return;
    }

    vorbis_decoder_t *decoder = malloc(sizeof(*decoder));
    if (!decoder) {
        goto error_return;
    }
    decoder->pcm = NULL;
    decoder->num_samples = 0;

    decoder->page_number = 0;
    decoder->packet = id_header;
    decoder->packet_len = id_length;
    decoder->next_packet = setup_header;
    decoder->next_packet_len = setup_length;
    gen_page_header(decoder, 1);
    decoder->nogg = vorbis_open_callbacks(
        ((const vorbis_callbacks_t){.read = nogg_read}), decoder, 0, NULL);
    if (!decoder->nogg) {
        goto error_free_decoder;
    }

    return decoder;

  error_free_decoder:
    free(decoder);
  error_return:
    return NULL;
}

/*-----------------------------------------------------------------------*/

void vorbis_decoder_destroy(vorbis_decoder_t *decoder)
{
    free(decoder->pcm);
    vorbis_close(decoder->nogg);
    free(decoder);
}

/*-----------------------------------------------------------------------*/

void vorbis_decoder_reset(vorbis_decoder_t *decoder)
{
    /* Nothing to do. */
}

/*-----------------------------------------------------------------------*/

int vorbis_decoder_decode(
    vorbis_decoder_t *decoder, const void *data, int length)
{
    decoder->packet = data;
    decoder->packet_len = length;
    gen_page_header(decoder, 0);

    const int channels = vorbis_channels(decoder->nogg);
    const int sample_size = 4 * channels;
    float read_buf[1024];
    int32_t new_samples;
    while ((new_samples = vorbis_read_float(
                decoder->nogg, read_buf, sizeof(read_buf) / sample_size,
                NULL)) > 0) {
        const int32_t total_samples = decoder->num_samples + new_samples;
        float *new_pcm = realloc(decoder->pcm, total_samples * sample_size);
        if (!new_pcm) {
            return 0;
        }
        memcpy(new_pcm + (decoder->num_samples * channels), read_buf,
               new_samples * sample_size);
        decoder->pcm = new_pcm;
        decoder->num_samples = total_samples;
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

int vorbis_decoder_available_samples(const vorbis_decoder_t *decoder)
{
    return decoder->num_samples;
}

/*-----------------------------------------------------------------------*/

int vorbis_decoder_get_pcm(
    vorbis_decoder_t *decoder, float *buffer, int samples)
{
    if (samples > decoder->num_samples) {
        samples = decoder->num_samples;
    }
    if (samples <= 0) {
        return 0;
    }

    const int channels = vorbis_channels(decoder->nogg);
    const int sample_size = 4 * channels;
    memcpy(buffer, decoder->pcm, samples * sample_size);
    decoder->num_samples -= samples;
    if (decoder->num_samples > 0) {
        memmove(decoder->pcm, decoder->pcm + (samples * channels),
                decoder->num_samples * sample_size);
        float *new_pcm =
            realloc(decoder->pcm, decoder->num_samples * sample_size);
        /* This realloc should never fail, but if it does, just leave the
         * larger buffer alone since it won't break anything. */
        if (new_pcm) {
            decoder->pcm = new_pcm;
        }
    } else {
        free(decoder->pcm);
        decoder->pcm = NULL;
    }

    return samples;
}

/*************************************************************************/
/*************************************************************************/

#endif  /* DECODE_AUDIO */
